#ifndef _XPS_LIB_H
#define _XPS_LIB_H

#include "portable.h"

#include <stdio.h>

#include <ac/string.h>
#include <ac/socket.h>

#include "ldap_pvt.h"
#include "lutil.h"
#include "slap.h"
#include "lutil_ldap.h"

#include <openssl/md5.h>

typedef unsigned char byte;

struct _typedesc; 

typedef char *( * _CPFPLUS )( char *, int, int, unsigned char **,
	int, struct _typedesc[], int, int, char *, Modifications ***, byte ** );

typedef struct _typedesc {
	char * name;
	int tag; // just for user specified tags
	int primitive; // 30 for SEQUENCE etc
	int tagOption; // IMPLICIT_TAG or EXPLICIT_TAG
	_CPFPLUS function;
	int optional;
	int Default;
	char *ObjectIdentifier;
	Modifications ***xps_modtail ;
} TypeDescription;

typedef struct _oid {
	int length;
	unsigned char *oidbytes;
} OID;

typedef struct _oidfunct {
	_CPFPLUS function;
	 char *oidString;
	 char *name;
} OidFunction;

typedef struct _oidref {
	char *name;
	OID  *oid;
	struct _oidref *next;
} OIDREF;

typedef struct _tagg {
	int flag; // implicit or explicit
	int value;// the actual tagoidbytes
#define IMPLICIT_TAG 1
#define EXPLICIT_TAG 0
} Tag;

typedef struct _asn1type {
	char name[100];
	int token;
} ASN1Type;

typedef struct _type {
	char name[100];
	ASN1Type aType;
	struct _type *next;
	Tag tag;
	int optional;// 1 if OPTIONAL 0 otherwise
	int Default;// 1 if this is a default type  
	char ObjectIdentifier[100];
    int value; // for NUMBER types & enumerations
    int Primitive; // 1 if a primitive 0 if constructed
    struct _type *typelist;// for the types in a sequence or choice
									// or SET OF, or SET
} Type;

OID *encodeOID (
	char *str );

char *newString (
	char *str);

char * decodeOCTET_STRING ( 
	char *name,
	int tag,
	int implicit,
	byte **P,
	int count,
	TypeDescription TD[],
	int optional,
	int Default,
	char *ObjectIdentifier,
	Modifications ***xps_modtail, byte **P_CRL  );

char *
decodeINTEGER (char *name, int tag, int implicit, byte ** P, int count,
	       TypeDescription TD[], int optional, int Default,
	       char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );

char *decode_5_1(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
		   
int xps_free_entries (
	Operation *o,
	Entry **e );

int xps_open_wal (
	FILE **wal,
	char **walname,
	struct berval dn );

int xps_close_wal (
	FILE *wal,
	char *walname );
	
int xps_dn2wal (
	FILE *wal,
	struct berval *dn );

int xps_entry2wal (
	FILE *wal,
	Entry *e );

char *xps_read_last_record (
	FILE *wal );
	
int xps_delete_last_record (
	FILE *wal,
	char *walname );
	
int xps_process_wal (
	Operation *op,
	SlapReply *rs,
	FILE *wal,
	char *walname );

int xps_add_mod (
	char *name,
	char *value,
	int len,
	Modifications ***xps_modtail,
	int flag );

void xps_add_oc (
	char		  *oc,
	char		  *abs_oc,
	Modifications	***modtail,
	struct berval	*x509_type,
	struct berval	*x509_value );

char *xps_find_attr(
	char *type, 
//	int flag,
	Modifications	*modlist );

int xps_create_dn(
	int				type,
	struct berval	*dn,
	struct berval	*xps_dn,
	Modifications	*xps_modlist );
	
int xps_x509attrs2entries(
	Operation *op,
	SlapReply *rs,
	Modifications	*mod,
	char *oc_type,
	char *abs_oc_type,
	int dnflag,
	struct berval dn,
	_CPFPLUS decodeFUNCTION,
	char *decodeName,
	Entry ***e,
	Entry ***e_revoked );

int xps_create_entries(
	Operation *op,
	SlapReply *rs,
	Modifications	**modtail,
	Entry	***xps_e_array,
	Entry	***xps_e_revoked,
	struct berval dn );

int xps_add_entries(
	Operation *op,
	SlapReply *rs,
	FILE *wal,
	Entry **e,
	char *walname );

int xps_send_search_entry( Operation *op, SlapReply *rs );

slap_response xps_send_search_entry;
	
int xps_create_filter(
 	Operation *op,	
	Filter **filt,
	struct berval *a_type,
	struct berval *a_value );
	
int xps_delete_entries ( Operation *op, SlapReply *rs, FILE *wal, char *walname );

int xps_create_entries_for_modify(
	Operation *op,
	SlapReply *rs,
	Modifications	**modtail,
	Entry	***xps_e_array,
	Entry	***xps_e_revoked,
	struct berval dn );
	
int xps_process_modify (
	Operation *op,
	SlapReply *rs,
	Modifications	*mod,
	char *oc_type,
	char *abs_oc_type,
	int dnflag,
	struct berval dn,
	_CPFPLUS decodeFUNCTION,
	char *decodeName,
	Entry ***e,
	Entry ***e_revoked );
	

int xps_get_level3_entries( Operation *xps_op,	SlapReply *rs );

char * decodeC(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeO(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeOu(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeCn(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeL(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeSt(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeE_mail(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeDc(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeAttributeCertificate(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeAttributeCertificateInfo(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_0(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeAttributes(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeAttCertVersion(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeHolder(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeObjectDigestInfo(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeAttCertIssuer(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeV2Form(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeIssuerSerial(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeAttCertValidityPeriod(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeAttributeTypeAndValue(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_1(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeAttribute(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeX520name(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeX520CommonName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeX520LocalityName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeX520StateOrProvinceName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeX520OrganizationName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeX520OrganizationalUnitName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeX520Title(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeX520dnQualifier(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeX520countryName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeX520SerialNumber(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeX520Pseudonym(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeX520DomainComponent(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeEmailAddress(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_2(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeRDNSequence(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeDistinguishedName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_3(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeRelativeDistinguishedName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeDirectoryString(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeCertificate(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeTBSCertificate(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeVersion(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeCertificateSerialNumber(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeValidity(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeTime(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeUniqueIdentifier(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeSubjectPublicKeyInfo(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_4(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeExtensions(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeExtension(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeCertificateList(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_5(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeTBSCertList(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeAlgorithmIdentifier(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeAuthorityKeyIdentifier(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeKeyIdentifier(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeSubjectKeyIdentifier(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeKeyUsage(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodePrivateKeyUsagePeriod(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_6(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeCertificatePolicies(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_7(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodePolicyInformation(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeCertPolicyId(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodePolicyQualifierInfo(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeUserNotice(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_8(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeNoticeReference(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeDisplayText(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_9(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodePolicyMappings(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeSubjectAltName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_10(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeGeneralNames(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeGeneralName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeAnotherName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeEDIPartyName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeIssuerAltName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_11(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeSubjectDirectoryAttributes(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeBasicConstraints(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeNameConstraints(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_12(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeGeneralSubtrees(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeGeneralSubtree(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeBaseDistance(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodePolicyConstraints(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeSkipCerts(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_13(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeCRLDistributionPoints(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeDistributionPoint(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeDistributionPointName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeReasonFlags(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_14(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeExtKeyUsageSyntax(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeKeyPurposeId(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeInhibitAnyPolicy(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeFreshestCRL(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_15(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeAuthorityInfoAccessSyntax(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeAccessDescription(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decode_16(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeSubjectInfoAccessSyntax(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_odtail, byte **P_CRL );
char * decodeCRLNumber(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeIssuingDistributionPoint(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeBaseCRLNumber(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeCRLReason(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeCertificateIssuer(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeHoldInstructionCode(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
char * decodeInvalidityDate(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL );
#endif /* _XPS_LIB_H */
